package java.time;

import java.time.chrono.ChronoLocalDate;
import java.time.chrono.ChronoLocalDateTime;
import java.io.ObjectInput;
import java.io.IOException;
import java.io.DataOutput;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.time.temporal.TemporalQueries;
import java.time.temporal.TemporalQuery;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalAmount;
import java.time.temporal.TemporalAdjuster;
import java.time.temporal.UnsupportedTemporalTypeException;
import java.time.temporal.ValueRange;
import java.time.temporal.TemporalUnit;
import java.time.temporal.TemporalField;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoField;
import java.time.temporal.TemporalAccessor;
import java.time.zone.ZoneOffsetTransition;
import java.util.List;
import java.time.zone.ZoneRules;
import java.util.Objects;
import java.io.Serializable;
import java.time.chrono.ChronoZonedDateTime;
import java.time.temporal.Temporal;

public final class ZonedDateTime implements Temporal, ChronoZonedDateTime<LocalDate>, Serializable
{
    private static final long serialVersionUID = -6260982410461394882L;
    private final LocalDateTime dateTime;
    private final ZoneOffset offset;
    private final ZoneId zone;
    
    public static ZonedDateTime now() {
        return now(Clock.systemDefaultZone());
    }
    
    public static ZonedDateTime now(ZoneId zoneId) {
        return now(Clock.system(zoneId));
    }
    
    public static ZonedDateTime now(Clock obj) {
        Objects.requireNonNull(obj, "clock");
        return ofInstant(obj.instant(), obj.getZone());
    }
    
    public static ZonedDateTime of(LocalDate localDate, LocalTime localTime, ZoneId zoneId) {
        return of(LocalDateTime.of(localDate, localTime), zoneId);
    }
    
    public static ZonedDateTime of(LocalDateTime localDateTime, ZoneId zoneId) {
        return ofLocal(localDateTime, zoneId, null);
    }
    
    public static ZonedDateTime of(int n, int n2, int n3, int n4, int n5, int n6, int n7, ZoneId zoneId) {
        return ofLocal(LocalDateTime.of(n, n2, n3, n4, n5, n6, n7), zoneId, null);
    }
    
    public static ZonedDateTime ofLocal(LocalDateTime plusSeconds, ZoneId obj, ZoneOffset zoneOffset) {
        Objects.requireNonNull(plusSeconds, "localDateTime");
        Objects.requireNonNull(obj, "zone");
        if ((obj instanceof ZoneOffset)) {
            return new ZonedDateTime(plusSeconds, (ZoneOffset)obj, obj);
        }
        ZoneRules rules = obj.getRules();
        List<ZoneOffset> validOffsets = rules.getValidOffsets(plusSeconds);
        ZoneOffset offsetAfter;
        if (validOffsets.size() == 1) {
            offsetAfter = (ZoneOffset)validOffsets.get(0);
        }
        else if (validOffsets.size() == 0) {
            ZoneOffsetTransition transition = rules.getTransition(plusSeconds);
            plusSeconds = plusSeconds.plusSeconds(transition.getDuration().getSeconds());
            offsetAfter = transition.getOffsetAfter();
        }
        else if (zoneOffset != null && validOffsets.contains(zoneOffset)) {
            offsetAfter = zoneOffset;
        }
        else {
            offsetAfter = (ZoneOffset)Objects.requireNonNull(validOffsets.get(0), "offset");
        }
        return new ZonedDateTime(plusSeconds, offsetAfter, obj);
    }
    
    public static ZonedDateTime ofInstant(Instant obj, ZoneId obj2) {
        Objects.requireNonNull(obj, "instant");
        Objects.requireNonNull(obj2, "zone");
        return create(obj.getEpochSecond(), obj.getNano(), obj2);
    }
    
    public static ZonedDateTime ofInstant(LocalDateTime localDateTime, ZoneOffset offset, ZoneId obj) {
        Objects.requireNonNull(localDateTime, "localDateTime");
        Objects.requireNonNull(offset, "offset");
        Objects.requireNonNull(obj, "zone");
        if (obj.getRules().isValidOffset(localDateTime, offset)) {
            return new ZonedDateTime(localDateTime, offset, obj);
        }
        return create(localDateTime.toEpochSecond(offset), localDateTime.getNano(), obj);
    }
    
    private static ZonedDateTime create(long n, int n2, ZoneId zoneId) {
        ZoneOffset offset = zoneId.getRules().getOffset(Instant.ofEpochSecond(n, (long)n2));
        return new ZonedDateTime(LocalDateTime.ofEpochSecond(n, n2, offset), offset, zoneId);
    }
    
    public static ZonedDateTime ofStrict(LocalDateTime obj, ZoneOffset obj2, ZoneId obj3) {
        Objects.requireNonNull(obj, "localDateTime");
        Objects.requireNonNull(obj2, "offset");
        Objects.requireNonNull(obj3, "zone");
        ZoneRules rules = obj3.getRules();
        if (!rules.isValidOffset(obj, obj2)) {
            ZoneOffsetTransition transition = rules.getTransition(obj);
            if (transition != null && transition.isGap()) {
                throw new DateTimeException("LocalDateTime '" + obj + "' does not exist in zone '" + obj3 + "' due to a gap in the local time-line, typically caused by daylight savings");
            }
            throw new DateTimeException("ZoneOffset '" + obj2 + "' is not valid for LocalDateTime '" + obj + "' in zone '" + obj3 + "'");
        }
        else {
            return new ZonedDateTime(obj, obj2, obj3);
        }
    }
    
    private static ZonedDateTime ofLenient(LocalDateTime obj, ZoneOffset obj2, ZoneId obj3) {
        Objects.requireNonNull(obj, "localDateTime");
        Objects.requireNonNull(obj2, "offset");
        Objects.requireNonNull(obj3, "zone");
        if (obj3 instanceof ZoneOffset && !obj2.equals(obj3)) {
            throw new IllegalArgumentException("ZoneId must match ZoneOffset");
        }
        return new ZonedDateTime(obj, obj2, obj3);
    }
    
    public static ZonedDateTime from(TemporalAccessor obj) {
        if ((obj instanceof ZonedDateTime)) {
            return (ZonedDateTime)obj;
        }
        try {
            ZoneId from = ZoneId.from(obj);
            if (obj.isSupported(ChronoField.INSTANT_SECONDS)) {
                return create(obj.getLong(ChronoField.INSTANT_SECONDS), obj.get(ChronoField.NANO_OF_SECOND), from);
            }
            return of(LocalDate.from(obj), LocalTime.from(obj), from);
        }
        catch (DateTimeException ex) {
            throw new DateTimeException("Unable to obtain ZonedDateTime from TemporalAccessor: " + obj + " of type " + obj.getClass().getName(), ex);
        }
    }
    
    public static ZonedDateTime parse(CharSequence charSequence) {
        return parse(charSequence, DateTimeFormatter.ISO_ZONED_DATE_TIME);
    }
    
    public static ZonedDateTime parse(CharSequence text, DateTimeFormatter obj) {
        Objects.requireNonNull(obj, "formatter");
        return (ZonedDateTime)obj.parse(text, ZonedDateTime::from);
    }
    
    private ZonedDateTime(LocalDateTime dateTime, ZoneOffset offset, ZoneId zone) {
        this.dateTime = dateTime;
        this.offset = offset;
        this.zone = zone;
    }
    
    private ZonedDateTime resolveLocal(LocalDateTime localDateTime) {
        return ofLocal(localDateTime, this.zone, this.offset);
    }
    
    private ZonedDateTime resolveInstant(LocalDateTime localDateTime) {
        return ofInstant(localDateTime, this.offset, this.zone);
    }
    
    private ZonedDateTime resolveOffset(ZoneOffset offset) {
        if (!offset.equals(this.offset) && this.zone.getRules().isValidOffset(this.dateTime, offset)) {
            return new ZonedDateTime(this.dateTime, offset, this.zone);
        }
        return this;
    }
    
    @Override
    public boolean isSupported(TemporalField temporalField) {
        return temporalField instanceof ChronoField || (temporalField != null && temporalField.isSupportedBy(this));
    }
    
    @Override
    public boolean isSupported(TemporalUnit unit) {
        return super.isSupported(unit);
    }
    
    @Override
    public ValueRange range(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            if (temporalField == ChronoField.INSTANT_SECONDS || temporalField == ChronoField.OFFSET_SECONDS) {
                return temporalField.range();
            }
            return this.dateTime.range(temporalField);
        }
        else {
            return temporalField.rangeRefinedBy(this);
        }
    }
    
    @Override
    public int get(TemporalField field) {
        if ((field instanceof ChronoField)) {
            switch ((ChronoField)field) {
                case INSTANT_SECONDS: {
                    throw new UnsupportedTemporalTypeException("Invalid field 'InstantSeconds' for get() method, use getLong() instead");
                }
                case OFFSET_SECONDS: {
                    return this.getOffset().getTotalSeconds();
                }
                default: {
                    return this.dateTime.get(field);
                }
            }
        }
        else {
            return super.get(field);
        }
    }
    
    @Override
    public long getLong(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            switch ((ChronoField)temporalField) {
                case INSTANT_SECONDS: {
                    return this.toEpochSecond();
                }
                case OFFSET_SECONDS: {
                    return (long)this.getOffset().getTotalSeconds();
                }
                default: {
                    return this.dateTime.getLong(temporalField);
                }
            }
        }
        else {
            return temporalField.getFrom(this);
        }
    }
    
    @Override
    public ZoneOffset getOffset() {
        return this.offset;
    }
    
    @Override
    public ZonedDateTime withEarlierOffsetAtOverlap() {
        ZoneOffsetTransition transition = this.getZone().getRules().getTransition(this.dateTime);
        if (transition != null && transition.isOverlap()) {
            ZoneOffset offsetBefore = transition.getOffsetBefore();
            if (!offsetBefore.equals(this.offset)) {
                return new ZonedDateTime(this.dateTime, offsetBefore, this.zone);
            }
        }
        return this;
    }
    
    @Override
    public ZonedDateTime withLaterOffsetAtOverlap() {
        ZoneOffsetTransition transition = this.getZone().getRules().getTransition(this.toLocalDateTime());
        if (transition != null) {
            ZoneOffset offsetAfter = transition.getOffsetAfter();
            if (!offsetAfter.equals(this.offset)) {
                return new ZonedDateTime(this.dateTime, offsetAfter, this.zone);
            }
        }
        return this;
    }
    
    @Override
    public ZoneId getZone() {
        return this.zone;
    }
    
    @Override
    public ZonedDateTime withZoneSameLocal(ZoneId obj) {
        Objects.requireNonNull(obj, "zone");
        return this.zone.equals(obj) ? this : ofLocal(this.dateTime, obj, this.offset);
    }
    
    @Override
    public ZonedDateTime withZoneSameInstant(ZoneId obj) {
        Objects.requireNonNull(obj, "zone");
        return this.zone.equals(obj) ? this : create(this.dateTime.toEpochSecond(this.offset), this.dateTime.getNano(), obj);
    }
    
    public ZonedDateTime withFixedOffsetZone() {
        return this.zone.equals(this.offset) ? this : new ZonedDateTime(this.dateTime, this.offset, this.offset);
    }
    
    @Override
    public LocalDateTime toLocalDateTime() {
        return this.dateTime;
    }
    
    @Override
    public LocalDate toLocalDate() {
        return this.dateTime.toLocalDate();
    }
    
    public int getYear() {
        return this.dateTime.getYear();
    }
    
    public int getMonthValue() {
        return this.dateTime.getMonthValue();
    }
    
    public Month getMonth() {
        return this.dateTime.getMonth();
    }
    
    public int getDayOfMonth() {
        return this.dateTime.getDayOfMonth();
    }
    
    public int getDayOfYear() {
        return this.dateTime.getDayOfYear();
    }
    
    public DayOfWeek getDayOfWeek() {
        return this.dateTime.getDayOfWeek();
    }
    
    @Override
    public LocalTime toLocalTime() {
        return this.dateTime.toLocalTime();
    }
    
    public int getHour() {
        return this.dateTime.getHour();
    }
    
    public int getMinute() {
        return this.dateTime.getMinute();
    }
    
    public int getSecond() {
        return this.dateTime.getSecond();
    }
    
    public int getNano() {
        return this.dateTime.getNano();
    }
    
    @Override
    public ZonedDateTime with(TemporalAdjuster temporalAdjuster) {
        if ((temporalAdjuster instanceof LocalDate)) {
            return this.resolveLocal(LocalDateTime.of((LocalDate)temporalAdjuster, this.dateTime.toLocalTime()));
        }
        if ((temporalAdjuster instanceof LocalTime)) {
            return this.resolveLocal(LocalDateTime.of(this.dateTime.toLocalDate(), (LocalTime)temporalAdjuster));
        }
        if ((temporalAdjuster instanceof LocalDateTime)) {
            return this.resolveLocal((LocalDateTime)temporalAdjuster);
        }
        if ((temporalAdjuster instanceof OffsetDateTime)) {
            OffsetDateTime offsetDateTime = (OffsetDateTime)temporalAdjuster;
            return ofLocal(offsetDateTime.toLocalDateTime(), this.zone, offsetDateTime.getOffset());
        }
        if ((temporalAdjuster instanceof Instant)) {
            Instant instant = (Instant)temporalAdjuster;
            return create(instant.getEpochSecond(), instant.getNano(), this.zone);
        }
        if ((temporalAdjuster instanceof ZoneOffset)) {
            return this.resolveOffset((ZoneOffset)temporalAdjuster);
        }
        return (ZonedDateTime)temporalAdjuster.adjustInto(this);
    }
    
    @Override
    public ZonedDateTime with(TemporalField temporalField, long value) {
        if ((temporalField instanceof ChronoField)) {
            ChronoField chronoField = (ChronoField)temporalField;
            switch (chronoField) {
                case INSTANT_SECONDS: {
                    return create(value, this.getNano(), this.zone);
                }
                case OFFSET_SECONDS: {
                    return this.resolveOffset(ZoneOffset.ofTotalSeconds(chronoField.checkValidIntValue(value)));
                }
                default: {
                    return this.resolveLocal(this.dateTime.with(temporalField, value));
                }
            }
        }
        else {
            return (ZonedDateTime)temporalField.adjustInto(this, value);
        }
    }
    
    public ZonedDateTime withYear(int n) {
        return this.resolveLocal(this.dateTime.withYear(n));
    }
    
    public ZonedDateTime withMonth(int n) {
        return this.resolveLocal(this.dateTime.withMonth(n));
    }
    
    public ZonedDateTime withDayOfMonth(int n) {
        return this.resolveLocal(this.dateTime.withDayOfMonth(n));
    }
    
    public ZonedDateTime withDayOfYear(int n) {
        return this.resolveLocal(this.dateTime.withDayOfYear(n));
    }
    
    public ZonedDateTime withHour(int n) {
        return this.resolveLocal(this.dateTime.withHour(n));
    }
    
    public ZonedDateTime withMinute(int n) {
        return this.resolveLocal(this.dateTime.withMinute(n));
    }
    
    public ZonedDateTime withSecond(int n) {
        return this.resolveLocal(this.dateTime.withSecond(n));
    }
    
    public ZonedDateTime withNano(int n) {
        return this.resolveLocal(this.dateTime.withNano(n));
    }
    
    public ZonedDateTime truncatedTo(TemporalUnit temporalUnit) {
        return this.resolveLocal(this.dateTime.truncatedTo(temporalUnit));
    }
    
    @Override
    public ZonedDateTime plus(TemporalAmount obj) {
        if ((obj instanceof Period)) {
            return this.resolveLocal(this.dateTime.plus((TemporalAmount)obj));
        }
        Objects.requireNonNull(obj, "amountToAdd");
        return (ZonedDateTime)obj.addTo(this);
    }
    
    @Override
    public ZonedDateTime plus(long n, TemporalUnit temporalUnit) {
        if ((temporalUnit instanceof ChronoUnit)) {
            if (temporalUnit.isDateBased()) {
                return this.resolveLocal(this.dateTime.plus(n, temporalUnit));
            }
            return this.resolveInstant(this.dateTime.plus(n, temporalUnit));
        }
        else {
            return (ZonedDateTime)temporalUnit.addTo(this, n);
        }
    }
    
    public ZonedDateTime plusYears(long n) {
        return this.resolveLocal(this.dateTime.plusYears(n));
    }
    
    public ZonedDateTime plusMonths(long n) {
        return this.resolveLocal(this.dateTime.plusMonths(n));
    }
    
    public ZonedDateTime plusWeeks(long n) {
        return this.resolveLocal(this.dateTime.plusWeeks(n));
    }
    
    public ZonedDateTime plusDays(long n) {
        return this.resolveLocal(this.dateTime.plusDays(n));
    }
    
    public ZonedDateTime plusHours(long n) {
        return this.resolveInstant(this.dateTime.plusHours(n));
    }
    
    public ZonedDateTime plusMinutes(long n) {
        return this.resolveInstant(this.dateTime.plusMinutes(n));
    }
    
    public ZonedDateTime plusSeconds(long n) {
        return this.resolveInstant(this.dateTime.plusSeconds(n));
    }
    
    public ZonedDateTime plusNanos(long n) {
        return this.resolveInstant(this.dateTime.plusNanos(n));
    }
    
    @Override
    public ZonedDateTime minus(TemporalAmount obj) {
        if ((obj instanceof Period)) {
            return this.resolveLocal(this.dateTime.minus((TemporalAmount)obj));
        }
        Objects.requireNonNull(obj, "amountToSubtract");
        return (ZonedDateTime)obj.subtractFrom(this);
    }
    
    @Override
    public ZonedDateTime minus(long n, TemporalUnit temporalUnit) {
        return (n == Long.MIN_VALUE) ? this.plus(Long.MAX_VALUE, temporalUnit).plus(1L, temporalUnit) : this.plus(-n, temporalUnit);
    }
    
    public ZonedDateTime minusYears(long n) {
        return (n == Long.MIN_VALUE) ? this.plusYears(Long.MAX_VALUE).plusYears(1L) : this.plusYears(-n);
    }
    
    public ZonedDateTime minusMonths(long n) {
        return (n == Long.MIN_VALUE) ? this.plusMonths(Long.MAX_VALUE).plusMonths(1L) : this.plusMonths(-n);
    }
    
    public ZonedDateTime minusWeeks(long n) {
        return (n == Long.MIN_VALUE) ? this.plusWeeks(Long.MAX_VALUE).plusWeeks(1L) : this.plusWeeks(-n);
    }
    
    public ZonedDateTime minusDays(long n) {
        return (n == Long.MIN_VALUE) ? this.plusDays(Long.MAX_VALUE).plusDays(1L) : this.plusDays(-n);
    }
    
    public ZonedDateTime minusHours(long n) {
        return (n == Long.MIN_VALUE) ? this.plusHours(Long.MAX_VALUE).plusHours(1L) : this.plusHours(-n);
    }
    
    public ZonedDateTime minusMinutes(long n) {
        return (n == Long.MIN_VALUE) ? this.plusMinutes(Long.MAX_VALUE).plusMinutes(1L) : this.plusMinutes(-n);
    }
    
    public ZonedDateTime minusSeconds(long n) {
        return (n == Long.MIN_VALUE) ? this.plusSeconds(Long.MAX_VALUE).plusSeconds(1L) : this.plusSeconds(-n);
    }
    
    public ZonedDateTime minusNanos(long n) {
        return (n == Long.MIN_VALUE) ? this.plusNanos(Long.MAX_VALUE).plusNanos(1L) : this.plusNanos(-n);
    }
    
    @Override
    public <R> R query(TemporalQuery<R> query) {
        if (query == TemporalQueries.localDate()) {
            return this.toLocalDate();
        }
        return super.query(query);
    }
    
    @Override
    public long until(Temporal temporal, TemporalUnit temporalUnit) {
        ZonedDateTime from = from(temporal);
        if ((temporalUnit instanceof ChronoUnit)) {
            ZonedDateTime withZoneSameInstant = from.withZoneSameInstant(this.zone);
            if (temporalUnit.isDateBased()) {
                return this.dateTime.until(withZoneSameInstant.dateTime, temporalUnit);
            }
            return this.toOffsetDateTime().until(withZoneSameInstant.toOffsetDateTime(), temporalUnit);
        }
        else {
            return temporalUnit.between(this, from);
        }
    }
    
    @Override
    public String format(DateTimeFormatter obj) {
        Objects.requireNonNull(obj, "formatter");
        return obj.format(this);
    }
    
    public OffsetDateTime toOffsetDateTime() {
        return OffsetDateTime.of(this.dateTime, this.offset);
    }
    
    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if ((o instanceof ZonedDateTime)) {
            ZonedDateTime zonedDateTime = (ZonedDateTime)o;
            return this.dateTime.equals(zonedDateTime.dateTime) && this.offset.equals(zonedDateTime.offset) && this.zone.equals(zonedDateTime.zone);
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return this.dateTime.hashCode() ^ this.offset.hashCode() ^ Integer.rotateLeft(this.zone.hashCode(), 3);
    }
    
    @Override
    public String toString() {
        String str = this.dateTime.toString() + this.offset.toString();
        if (this.offset != this.zone) {
            str = str + 91 + this.zone.toString() + 93;
            return str;
        }
        return str;
    }
    
    private Object writeReplace() {
        return new Ser((byte)6, this);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    void writeExternal(DataOutput dataOutput) throws IOException {
        this.dateTime.writeExternal(dataOutput);
        this.offset.writeExternal(dataOutput);
        this.zone.write(dataOutput);
    }
    
    static ZonedDateTime readExternal(ObjectInput objectInput) throws IOException, ClassNotFoundException {
        return ofLenient(LocalDateTime.readExternal(objectInput), ZoneOffset.readExternal(objectInput), (ZoneId)Ser.read(objectInput));
    }
}
